/******************************************************************/
/*									                              */
/*      Description : SVD process                                 */
/*      Making Date : 2010. 05. 06					              */
/*      Affiliate   : Machine Learning Lab in SoongSil Univ       */
/*      Developer   : Sungin Park(rfchief@gmail.com)              */
/******************************************************************/

#include <stdio.h>
#include <cuda.h>
#include <culapack.h>
#include "CalEvalues.h"
#include "FileReader.h"
#include "FileWriter.h"
#include "utilFunc.h"

int calEvaluesProcess(struct paramsStr params, int divCnt, int divMax){
	int size_x           = params.size;
	int size_y           = params.size;	
	char labelChk        = params.header;		
	int p                = divCnt;
	int s_f              = 0, s_l = 0;	
	int dSize_x_l        = 0;
	int dSize_x          = 0;
	unsigned long startTime = 0l, endTime = 0l;
	int t = 0, diag = 0, idx = 0, g = 0, i = 0, j = 0;
	float *inputData     = (float *)malloc(size_x * size_y * sizeof(float));
	int *divDataIdx;
	float *eValueOfData;
	float **divData;
	char  **labelData; 
	float **eValueList;   
	char fPath[1100];
	char logMsg[1100];
	char *outName = params.outName;
	char *f_Name  = params.inputName;		
	culaStatus cStatus;
	
	sprintf(fPath, "%s_Log.txt", outName);	
	sprintf(logMsg, "Divide count : %d", p);
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);	
	sprintf(logMsg, "Calculate eigenvalues process starts\n");
	writeLogFile(fPath, logMsg, 0);	

	sprintf(fPath, "%s", f_Name); 
	if(labelChk == 'y' || labelChk == 'Y'){              
	    labelData = (char **)malloc(size_x*sizeof(char*));

		if(GetData(fPath, inputData, labelData, size_x)) 
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			errorExit(0);
		}	

		free(labelData);
	
	}else{                          //label  

		if(GetData2(fPath, inputData, size_x)) //file read  
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{ 
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			errorExit(0);
		}

	}//labelChk	
	
	startTime = clock();

	/* Initialize CULA */        
    cStatus = culaInitialize();
    checkStatus(cStatus, params);

	if(p > 0){
		dSize_x      = makingDivide(size_x, &p, divMax, &dSize_x_l, &s_f, &s_l);			
		divDataIdx   = (int *)malloc(dSize_x * p * sizeof(int));
		divData      = (float **)malloc(p * sizeof(float*));
		eValueList   = (float **)malloc(p * sizeof(float*));    

		for(i = 0; i < p; i++){
			if(i == (p-1)){
				divData[i] = (float *)malloc(dSize_x_l * dSize_x_l * sizeof(float));							
			}else{
				divData[i] = (float *)malloc(dSize_x * dSize_x * sizeof(float));				
			}
		}	

		for(i = 0; i < p; i++){
			if(i == (p-1)){
				eValueList[i] = (float *)malloc(dSize_x_l * sizeof(float));
			}else{
				eValueList[i] = (float *)malloc(dSize_x * sizeof(float));
			}			
		}


		///////////////////////////////////////////////////////////////////////
		///////////Data Decomposition Part
		///////////////////////////////////////////////////////////////////////
		for(i = 0; i < size_x; i++)
		{
			if(g == (p-1)){
				if(i % dSize_x_l == 0){
					divDataIdx[i] = (i * size_x) + i;			
					diag = divDataIdx[i];
				}else{
					divDataIdx[i] = diag + size_x;
					diag = divDataIdx[i];
				}
			}else{
				if(i % dSize_x == 0){
					divDataIdx[i] = (i * size_x) + i;			
					diag = divDataIdx[i];
				}else{
					divDataIdx[i] = diag + size_x;
					diag = divDataIdx[i];
				}
			}

			g++;
		}
	    
		for(i = 0;i < p;i++)
		{		
			diag = 0;

			if(i == (p-1)){		
				t = dSize_x_l;
			}else{
				t = dSize_x;
			}

			for(j = 0; j < t * t; j++)
			{			
				if(diag == t)
				{
					diag = 0;
					idx++;
				}

				divData[i][j] = inputData[divDataIdx[idx] + diag];			
				
				diag++;			
			}

			idx++;
		}

		///////////////////////////////////////////////////////////////////////
		///////////Data Decomposition Part end
		///////////////////////////////////////////////////////////////////////
		sprintf(fPath, "%s_Log.txt", outName);
		sprintf(logMsg, "SVD process starts");
		writeLogFile(fPath, logMsg, 0);

		for(i = 0; i<p; i++)
		{
			if(i == (p-1)){
				eValueList[i] = SVD(dSize_x_l, dSize_x_l, divData[i], params);       
			}else{
				eValueList[i] = SVD(dSize_x, dSize_x, divData[i], params);           
			}		
		}

		sprintf(logMsg, "SVD process sucess");
		writeLogFile(fPath, logMsg, 0);

		sprintf(fPath, "%s_EigenValues.txt", outName); 	
		if(writeeValueFileList(fPath, dSize_x, dSize_x_l, p, eValueList)){
			printf("Result file write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);
		}else{
			printf("Result file write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);
			
			free(inputData);
			free(divDataIdx);
			free(divData);
			free(eValueList);
			
			culaShutdown();

			errorExit(4);
		}
		
		free(divDataIdx);
		free(divData);
		free(eValueList);

	}else{
		sprintf(fPath, "%s_Log.txt", outName);
		sprintf(logMsg, "SVD process starts");
		writeLogFile(fPath, logMsg, 0);
		
		printf("SVD Processing\n");
		eValueOfData = SVD(size_x, size_x, inputData, params);
		
		sprintf(logMsg, "SVD process sucess");
		writeLogFile(fPath, logMsg, 0);

		sprintf(fPath, "%s_EigenValues.txt", outName); 	
		if(writeeValueFile(fPath, size_x, eValueOfData)){
			printf("Result file write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);

		}else{
			printf("Result file write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);
			
			free(inputData);
			free(eValueOfData);
			culaShutdown();

			errorExit(4);
		}
		
		free(eValueOfData);
	}//if(p>0)

	endTime = clock();
	printf("Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	sprintf(logMsg, "Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	writeLogFile(fPath, logMsg, 0);		
	sprintf(logMsg, "Calculate eigenvalues process starts");
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);

	free(inputData);	
	culaShutdown();
	
	return 0;
}

//Singular Value Decomposition
float* SVD (int pSize_x, int pSize_y, float *inputData, struct paramsStr params)
{	
	///////////////////////////////////////////////////////
	//////Initialize part
	////////////////////////////////////////////////////////
    int size_x = pSize_x;                                             //row
	int size_y = pSize_y;                                             //col	
	char jobu = 'A';
    char jobvt = 'A';
	float *h_PMatrix    = inputData;                                                //Input Matrix
    float* S = NULL;
    float* U = NULL;
    float* VT = NULL;
	culaStatus cStatus;
    	
	S       = (float*)malloc(imin(size_x,size_y)*sizeof(float));                //eigenvalues of SVD    
    U       = (float*)malloc(size_x*size_x*sizeof(float));                          //eigenvectors of SVD	
    VT      = (float*)malloc(size_x*size_x*sizeof(float));
	    
	///////////////////
	////SVD Process
	////////////////////
    
	//compute SVD in CULA
    cStatus = culaSgesvd(jobu, jobvt, size_x, size_x, h_PMatrix, size_x, S, U, size_x, VT, size_x);
    checkStatus(cStatus, params);
    //SVD Process end
	
	///////////////////
	////SVD Process end
	////////////////////

    /* Clean up workspace, input, and output memory allocations */	
	free(VT);
    free(U);
    
	return S;
}//SVD

