/******************************************************************/
/*									                              */
/*      Description : Sampling process                            */
/*      Making Date : 2010. 05. 06					              */
/*      Affiliate   : Machine Learning Lab in SoongSil Univ       */
/*      Developer   : Sungin Park(rfchief@gmail.com)              */
/******************************************************************/

#include <time.h>
#include <math.h>
#include <cuda.h>
#include <cublas.h>
#include <culapack.h>
#include "sampling.h"
#include "dataStruct.h"
#include "qSort.h"
#include "FileReader.h"
#include "FileWriter.h"
#include "utilFunc.h"
#include "cMDS.h"


int runMDS(struct paramsStr params){
	//parameter setting	
	int size_x           = params.size;
	int size_y           = params.size;
	int dim              = params.dimension;	
	char labelChk        = params.header;			
	unsigned long startTime = 0l, endTime = 0l;	
	float *inputData     = (float *)malloc(size_x*size_y*sizeof(float));	
	float *resultOfCMDS;     
	char  **labelData;
	char fPath[1100];
	char logMsg[1100];
	char *outName = params.outName;
	char *f_Name  = params.inputName;	
	cublasStatus status;
	culaStatus cStatus;
	
	printf("\n\nClassical MDS Process starts\n");
	sprintf(fPath, "%s_Log.txt", outName);
	writeLogFile(fPath, logMsg, 2);
	sprintf(logMsg, "Classical MDS Process starts\n");
	writeLogFile(fPath, logMsg, 0);

	sprintf(fPath, "%s", f_Name); 
	if(labelChk == 'y' || labelChk == 'Y'){              
	    labelData = (char **)malloc(size_x*sizeof(char*));

		if(GetData(fPath, inputData, labelData, size_x)) 
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			errorExit(0);
		}	
	
	}else{                         

		if(GetData2(fPath, inputData, size_x)) 
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{ 
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			errorExit(0);
		}

	}//labelChk		

	startTime = clock();

	/* Initialize CUBLAS & CULA */    
    status = cublasInit();
    cublasCheckStatus(status, params);    
    cStatus = culaInitialize();
    checkStatus(cStatus, params);

	printf("MDS Processing\n");
	resultOfCMDS = cMDS(size_x, size_y, dim, inputData, params);
	
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "MDS process success");
	writeLogFile(fPath, logMsg, 0);

	sprintf(fPath, "%s_CFMDS.txt", outName); 	
	if(labelChk == 'y' || labelChk == 'Y'){
		if(writeResultFileH(fPath, size_x, dim, resultOfCMDS, labelData)){
			printf("File write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);

			free(labelData);
		}else{
			printf("File write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			free(labelData);
			free(inputData);
			free(resultOfCMDS);

			culaShutdown();
			cublasShutdown();
			errorExit(4);
		}
	}else{
		if(writeResultFile(fPath, size_x, dim, resultOfCMDS)){
			printf("File write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);
		}else{
			printf("File write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			free(inputData);
			free(resultOfCMDS);

			culaShutdown();
			cublasShutdown();
			errorExit(4);
		}
	}
		
	endTime = clock();
	printf("Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);	
	sprintf(logMsg, "Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	writeLogFile(fPath, logMsg, 0);		
	sprintf(logMsg, "Classical MDS Process success");
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);
	
	free(inputData);
	free(resultOfCMDS);     

	culaShutdown();
	cublasShutdown();

	return 0;
}//runMDS

int runRnd(struct paramsStr params, int divCnt, int divMax)
{
	int i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, t = 0, g = 0;	
	int size_x           = params.size;
	int size_y           = params.size;
	int dim              = params.dimension;			
	char labelChk        = params.header;		
	int p                = divCnt;
	int s_f              = 0, s_l = 0;	
	int dSize_x_l        = 0;
	int dSize_x          = makingDivide(size_x, &p, divMax, &dSize_x_l, &s_f, &s_l);		
	int s                = s_f;
	int mAlignSize       = (s_f * (p-1)) + s_l;
	int chkLoop          = 1;                                //1 : false. 0 : true    
	unsigned long startTime = 0l, endTime = 0l;	
    float *inputData     = (float *)malloc(size_x * size_y * sizeof(float));
	int *samplePointIdx  = (int *)malloc(mAlignSize *sizeof(int));
	int *samplePointMDS  = (int *)malloc(mAlignSize *sizeof(int));
	float *mAlign        = (float *)malloc(mAlignSize * mAlignSize * sizeof(float));	
	float *affine        = (float *)malloc(dim*dim*sizeof(float));                                      //affine Matrix
	float *mAlignMDS     = (float *)malloc(mAlignSize*dim*sizeof(float));                               //mAlign MDS Matrix
	float *mAlignMDSTemp = (float *)malloc(s*dim*sizeof(float));                                        
    float *divDataMDSTemp= (float *)malloc(s*dim*sizeof(float));                                        
	float *subMDScoor    = (float *)malloc( dSize_x * dim * sizeof(float) );                            
	float *MDScoor       = (float *)malloc( size_x * dim * sizeof(float) );                             
	int *divDataIdx      = (int *)malloc(dSize_x * p * sizeof(int));
	int *samplePointChk  = (int *)malloc(s * sizeof(int));												//Sampled point idx temp	 
	float **mAlignTemp   = (float **)malloc(mAlignSize * sizeof(float*));
	float **divData      = (float **)malloc(p * sizeof(float*));	  
	float **divDataMDS   = (float **)malloc(p*sizeof(float*));                                          //Decomposition Data MDS Matrix		
	struct divDataStr *divDataS = (struct divDataStr*)malloc(dSize_x * sizeof(struct divDataStr));	
	char  **labelData;
	char fPath[1100];
	char logMsg[1100];
	char *outName = params.outName;            
	char *f_Name  = params.inputName;	       
	float *d_data1, *d_data2, *d_result;	
	cublasStatus status;
	culaStatus cStatus;
	
	for(i = 0; i < mAlignSize; i++){
		mAlignTemp[i] = (float *)malloc(mAlignSize * sizeof(float));
	}

	for(i = 0; i < p; i++){
		if(i == (p-1)){
			divData[i] = (float *)malloc(dSize_x_l * dSize_x_l * sizeof(float));		
			divDataMDS[i]   = (float *)malloc( dSize_x_l * dim * sizeof(float) );		
		}else{
			divData[i] = (float *)malloc(dSize_x * dSize_x * sizeof(float));		
			divDataMDS[i]   = (float *)malloc( dSize_x * dim * sizeof(float) );		
		}
	}

	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Divide count : %d\tSamples count : %d", p, s);
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);

	printf("\n\nCFMDS Process starts\n\n");	
	sprintf(logMsg, "CFMDS Process starts\n");
	writeLogFile(fPath, logMsg, 0);
	sprintf(logMsg, "The sampling strategy is the Random");
	writeLogFile(fPath, logMsg, 0);

	sprintf(fPath, "%s", f_Name); 
	if(labelChk == 'y' || labelChk == 'Y'){              

		labelData = (char **)malloc(size_x*sizeof(char*));

		if(GetData(fPath, inputData, labelData, size_x)) 
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			errorExit(0);
		}	
	
	}else{   

		if(GetData2(fPath, inputData, size_x)) //file read  
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{ 
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			errorExit(0);
		}

	}//labelChk		

	//rnd seed
	srand(time(NULL));	
	i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, t = 0, g = 0;
    startTime = 0l, endTime = 0l;   
		
	startTime = clock();
	
	///////////////////////////////////////////////////////////////////////
	///////////Data Decomposition Part
	///////////////////////////////////////////////////////////////////////    
	for(i = 0; i < size_x; i++)
	{
		if(g == (p-1)){
			if(i % dSize_x_l == 0){
				divDataIdx[i] = (i * size_x) + i;			
				diag = divDataIdx[i];
			}else{
				divDataIdx[i] = diag + size_x;
				diag = divDataIdx[i];
			}
		}else{
			if(i % dSize_x == 0){
				divDataIdx[i] = (i * size_x) + i;			
				diag = divDataIdx[i];
			}else{
				divDataIdx[i] = diag + size_x;
				diag = divDataIdx[i];
			}
		}

		g++;
	}
	
	for(i = 0;i < p;i++)
	{
		diag = 0;

		if(i == (p-1)){
			t = dSize_x_l;
		}else{
			t = dSize_x;
		}

		for(j = 0; j < t * t; j++)
		{			
			if(diag == t)
			{
				diag = 0;
				idx++;
			}

			divData[i][j] = inputData[divDataIdx[idx] + diag];			
			
			diag++;			
		}
		idx++;		
	}

	///////////////////////////////////////////////////////////////////////
	///////////Data Decomposition Part end
	///////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////
	///////////Data Sampling Part
	///////////////////////////////////////////////////////////////////////		
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Sampling process starts");
	writeLogFile(fPath, logMsg, 0);
    i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, t = 0, g = 0, idx = 0;    
	
	for(i=0; i<p; i++)
	{
		if(i == (p-1)){
			s = s_l;
			t = dSize_x_l;
		}else{
			s = s_f;
			t = dSize_x;
		}

		for(j=0; j<s; j++)
		{
			chkLoop = 0;
			samplePointIdx[idx] = rand()%t;		
			samplePointChk[j] = samplePointIdx[idx];			
			
			for(k = 0; k < j; k++)
			{
				if(samplePointChk[k] == samplePointIdx[idx])
				{
					j = j - 1;
					chkLoop = 1;
				}
			}
            
			if(chkLoop == 0){
				samplePointMDS[idx] = samplePointIdx[idx];
				samplePointIdx[idx] = samplePointIdx[idx] + (i*dSize_x);
				idx++;
			}
		}//for j
	}//for i

	sprintf(fPath, "%s_Indexes.txt", outName); 
	if(writeDataFileInt(fPath, mAlignSize, samplePointIdx)){
		printf("Index file write success\n");
	}else{
		printf("Index file write error\n");
		
		sprintf(fPath, "%s_Log.txt", outName);
		sprintf(logMsg, "Sampling process error");
		writeLogFile(fPath, logMsg, 0);
		sprintf(logMsg, "Program terminated");
		writeLogFile(fPath, logMsg, 0);
		writeLogFile(fPath, logMsg, 2);

		if(labelChk == 'y' || labelChk == 'Y'){
			free(labelData);
		}
		
		free(inputData);	
		free(samplePointIdx);
		free(samplePointMDS);	
		free(mAlign);
		free(mAlignTemp);
		free(mAlignMDS);	
		free(mAlignMDSTemp);
		free(divDataMDSTemp);
		free(affine);
		free(divDataS);
		free(divData);
		free(divDataIdx);
		free(divDataMDS);
		free(subMDScoor);
		free(MDScoor);
		free(samplePointChk);

		errorExit(4);
	}
		
	diag = 0; k = 1;idx = 0; mdsIdx = 0;                  
	for(i = 0; i < mAlignSize; i++)
	{
		for(j = i; j < mAlignSize; j++)
		{			
			if(i == j) 
			{
				mAlignTemp[i][j] = 0.0f;				
			}else{				
				mAlignTemp[i][j] = inputData[samplePointIdx[i] * size_x + samplePointIdx[j]];				
				mAlignTemp[j][i] = mAlignTemp[i][j];
			}     			
		}		
	}    
    	
	idx = 0;
	for(i = 0; i < mAlignSize; i++)
	{
		for(j = 0; j < mAlignSize; j++)
		{
			mAlign[idx] = mAlignTemp[i][j];				
			idx++;
		}
	}        
    	
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Sampling process success");
	writeLogFile(fPath, logMsg, 0);
	///////////////////////////////////////////////////////////////////////
	///////////Data Sampling end
	///////////////////////////////////////////////////////////////////////

	/* Initialize CUBLAS & CULA */
    status = cublasInit();
    cublasCheckStatus(status, params);
    cStatus = culaInitialize();
    checkStatus(cStatus, params);

    ///////////////////////////////////////////////////////////////////////
	///////////Classical MDS Part
	///////////////////////////////////////////////////////////////////////	    
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "MDS process starts");
	writeLogFile(fPath, logMsg, 0);

	mAlignMDS = cMDS(mAlignSize, mAlignSize, dim, mAlign, params);                                         
		
	for(i = 0; i<p; i++)
	{
		if(i == (p-1)){
			divDataMDS[i] = cMDS(dSize_x_l, dSize_x_l, dim, divData[i], params);            
		}else{
			divDataMDS[i] = cMDS(dSize_x, dSize_x, dim, divData[i], params);            
		}		
	}
	 
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "MDS process success");
	writeLogFile(fPath, logMsg, 0);
	///////////////////////////////////////////////////////////////////////
	///////////Classical MDS Part end
	///////////////////////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////////////////////
	///////////Calculate Least Linear Regression & Affine mapping Part
	///////////////////////////////////////////////////////////////////////
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Affine mapping process starts");
	writeLogFile(fPath, logMsg, 0);

    status = cublasAlloc(dSize_x*dim, sizeof(float), (void**)&d_data1);                                              //dMDS
    cublasCheckStatus(status, params);
    status = cublasAlloc(dim*dim, sizeof(float), (void**)&d_data2);                                                  //affine
    cublasCheckStatus(status, params);
    status = cublasAlloc(dSize_x*dim, sizeof(float), (void**)&d_result);                                             //MDS coordinate
    cublasCheckStatus(status, params);
	
	i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, t = 0, g = 0, idx = 0;
	for(diag = 0; diag<p; diag++)
	{
		if(diag == (p-1)){
			
			s = s_l;			
			subMDScoor = (float *)malloc( dSize_x_l * dim * sizeof(float) );
			mAlignMDSTemp  = (float *)malloc(s_l*dim*sizeof(float));
			divDataMDSTemp = (float *)malloc(s_l*dim*sizeof(float));

			cublasFree(d_data1);
			cublasFree(d_data2);
			cublasFree(d_result);

			status = cublasAlloc(dSize_x_l*dim, sizeof(float), (void**)&d_data1);                                              //dMDS
			cublasCheckStatus(status, params);
			status = cublasAlloc(dim*dim, sizeof(float), (void**)&d_data2);                                                    //affine
			cublasCheckStatus(status, params);
			status = cublasAlloc(dSize_x_l*dim, sizeof(float), (void**)&d_result);                                             //MDS Coordinate
			cublasCheckStatus(status, params);
			
			for(idx = 0; idx<dim; idx++){				
				for(k = 0; k<s_l; k++){
					if(idx == 0){
						t = (idx * s_f * p) + k + (diag * s_f);
					}else{
						t =  mAlignSize + k + (diag * s_f);
					}					
					mAlignMDSTemp[idx * s_l + k] = mAlignMDS[t];                                        							
					divDataMDSTemp[idx * s_l + k] = divDataMDS[diag][idx * dSize_x_l + samplePointMDS[k + s_f * diag]];									
				}			
			}

			cStatus = culaSgels('N', s_l, dim, dim, divDataMDSTemp, s_l, mAlignMDSTemp, s_l);           
			checkStatus(cStatus, params);

			for( i = 0; i < dim; i++ ) {
				for( j = 0; j < dim; j++ ){
					affine[i*dim+j] = mAlignMDSTemp[i*s_l+j];
				}			
			}

			status = cublasSetVector(dSize_x_l*dim, sizeof(float), divDataMDS[diag], 1, d_data1, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dim*dim, sizeof(float), affine, 1, d_data2, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dSize_x_l*dim, sizeof(float), subMDScoor, 1, d_result, 1);
			cublasCheckStatus(status, params);
			cublasSgemm('n', 'n', dSize_x_l, dim, dim, 1.0f, d_data1, dSize_x_l, d_data2, dim, 0.0f, d_result, dSize_x_l);                 //dMDS * A
			cublasGetVector(dSize_x_l*dim, sizeof(float), d_result, 1, subMDScoor, 1);
						
			mdsIdx = 0;
			for(i = 0; i<dim; i++)
			{
				for( j = 0; j < dSize_x_l; j++ ){
					MDScoor[(i * size_x) + j + (diag * dSize_x)] = subMDScoor[i*dSize_x_l+j];
					mdsIdx++;
				}
			}

		}else{
			
			for(idx = 0; idx<dim; idx++){
				for(k = 0; k<s; k++){
					mAlignMDSTemp[idx * s + k] = mAlignMDS[(idx * s * p) + k + (diag * s)];                                       
					divDataMDSTemp[idx * s + k] = divDataMDS[diag][idx * dSize_x + samplePointMDS[k + s*diag]];				
				}			
			}

			cStatus = culaSgels('N', s, dim, dim, divDataMDSTemp, s, mAlignMDSTemp, s);                                           
			checkStatus(cStatus, params);

			for( i = 0; i < dim; i++ ) {
				for( j = 0; j < dim; j++ ){
					affine[i*dim+j] = mAlignMDSTemp[i*s+j];
				}			
			}

			status = cublasSetVector(dSize_x*dim, sizeof(float), divDataMDS[diag], 1, d_data1, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dim*dim, sizeof(float), affine, 1, d_data2, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dSize_x*dim, sizeof(float), subMDScoor, 1, d_result, 1);
			cublasCheckStatus(status, params);
			cublasSgemm('n', 'n', dSize_x, dim, dim, 1.0f, d_data1, dSize_x, d_data2, dim, 0.0f, d_result, dSize_x);                 //dMDS * A
			cublasGetVector(dSize_x*dim, sizeof(float), d_result, 1, subMDScoor, 1);
						
			mdsIdx = 0;
			for(i = 0; i<dim; i++)
			{
				for( j = 0; j < dSize_x; j++ ){
					MDScoor[(i * size_x) + j + (diag * dSize_x)] = subMDScoor[i*dSize_x+j];
					mdsIdx++;
				}
			}
		}//if(i == (p-1))

	}//diag
    		
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Affine mapping process success");
	writeLogFile(fPath, logMsg, 0);
	///////////////////////////////////////////////////////////////////////
	///////////Calculate Least Linear Regression Part
	///////////////////////////////////////////////////////////////////////

	sprintf(fPath, "%s_CFMDS.txt", outName); 	
	if(labelChk == 'y' || labelChk == 'Y'){
		if(writeResultFileH(fPath, size_x, dim, MDScoor, labelData)){
			printf("File write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);

			free(labelData);
		}else{
			printf("File write error\n");	
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			cublasFree(d_data1);	
			cublasFree(d_data2);
			cublasFree(d_result);
			
			free(labelData);
		    free(inputData);	
		    free(samplePointIdx);
			free(samplePointMDS);
			free(mAlign);
			free(mAlignTemp);
			free(mAlignMDS);	
			free(mAlignMDSTemp);
			free(divDataMDSTemp);
			free(affine);
			free(divDataS);
			free(divData);	
			free(divDataIdx);
			free(divDataMDS);			
			free(subMDScoor);
			free(MDScoor);	
			free(samplePointChk);	

			culaShutdown();
			cublasShutdown();
			errorExit(4);
		}
	}else{
		if(writeResultFile(fPath, size_x, dim, MDScoor)){
			printf("File write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);
		}else{
			printf("File write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			cublasFree(d_data1);	
			cublasFree(d_data2);
			cublasFree(d_result);
				
		    free(inputData);	
		    free(samplePointIdx);
			free(samplePointMDS);
			free(mAlign);
			free(mAlignTemp);
			free(mAlignMDS);	
			free(mAlignMDSTemp);
			free(divDataMDSTemp);
			free(affine);
			free(divDataS);
			free(divData);	
			free(divDataIdx);
			free(divDataMDS);			
			free(subMDScoor);
			free(MDScoor);	
			free(samplePointChk);	

			culaShutdown();
			cublasShutdown();
			errorExit(4);
		}
	}//if(label chk)
	
	endTime = clock();	
	printf("Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	sprintf(logMsg, "Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	writeLogFile(fPath, logMsg, 0);		
	sprintf(logMsg, "CFMDS Process success");
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);

	cublasFree(d_data1);	
	cublasFree(d_data2);
	cublasFree(d_result);
	
    free(inputData);	
    free(samplePointIdx);
	free(samplePointMDS);
	free(mAlign);	
	free(mAlignTemp);
	free(mAlignMDS);	
	free(mAlignMDSTemp);
	free(divDataMDSTemp);
	free(affine);
	free(divDataS);
	free(divData);	
	free(divDataIdx);
	free(divDataMDS);	
	free(subMDScoor);
	free(MDScoor);	
	free(samplePointChk);			

	culaShutdown();
	cublasShutdown();			

	return 0;
}

int runMaxMin(struct paramsStr params, int divCnt, int divMax){
	int i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, h = 0, t = 0, x = 0, g = 0;	
	int size_x           = params.size;
	int size_y           = params.size;
	int dim              = params.dimension;	
	char labelChk        = params.header;		
	int p                = divCnt;
	int s_f              = 0, s_l = 0;	
	int dSize_x_l        = 0;
	int dSize_x          = makingDivide(size_x, &p, divMax, &dSize_x_l, &s_f, &s_l);		
	int s                = s_f;
	int mAlignSize       = (s_f * (p-1)) + s_l;
	unsigned long startTime = 0l, endTime = 0l;	
    float *inputData     = (float *)malloc(size_x * size_y * sizeof(float));
	int *samplePointIdx  = (int *)malloc(mAlignSize *sizeof(int));
	int *samplePointMDS  = (int *)malloc(mAlignSize *sizeof(int));
	float *minJ          = (float *)malloc(dSize_x * sizeof(float));	
	float *mAlign        = (float *)malloc(mAlignSize * mAlignSize * sizeof(float));	
	float *affine        = (float *)malloc(dim*dim*sizeof(float));                                      //affine Matrix
	float *mAlignMDS     = (float *)malloc(mAlignSize*dim*sizeof(float));                               //mAlign MDS Matrix
	float *mAlignMDSTemp = (float *)malloc(s*dim*sizeof(float));                                        
    float *divDataMDSTemp= (float *)malloc(s*dim*sizeof(float));                                        
	float *subMDScoor    = (float *)malloc( dSize_x * dim * sizeof(float) );                            
	float *MDScoor       = (float *)malloc( size_x * dim * sizeof(float) );                             
	int *divDataIdx      = (int *)malloc(dSize_x * p * sizeof(int));
	int *samplePointChk  = (int *)malloc(s * sizeof(int));												//Sampled point idx temp	 
	float **mAlignTemp   = (float **)malloc(mAlignSize * sizeof(float*));
	float **divData      = (float **)malloc(p * sizeof(float*));	  
	float **divDataMDS   = (float **)malloc(p*sizeof(float*));                                          //Decomposition Data MDS Matrix		                      
	struct divDataStr *divDataS = (struct divDataStr*)malloc(dSize_x * sizeof(struct divDataStr));	
	char  **labelData;
	char fPath[1100];
	char logMsg[1100];
	char *outName = params.outName;           
	char *f_Name  = params.inputName;	      
	float *d_data1, *d_data2, *d_result;	
	cublasStatus status;
	culaStatus cStatus;	
	
	for(i = 0; i < mAlignSize; i++){
		mAlignTemp[i] = (float *)malloc(mAlignSize * sizeof(float));
	}

	for(i = 0; i < p; i++){
		if(i == (p-1)){
			divData[i] = (float *)malloc(dSize_x_l * dSize_x_l * sizeof(float));		
		}else{
			divData[i] = (float *)malloc(dSize_x * dSize_x * sizeof(float));		
		}
	}
	
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Divide count : %d\tSamples count : %d", p, s);
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);

	printf("\n\nCFMDS Process starts\n\n");	
	sprintf(logMsg, "CFMDS Process starts\n");
	writeLogFile(fPath, logMsg, 0);
	sprintf(logMsg, "The sampling strategy is the Maxmin");
	writeLogFile(fPath, logMsg, 0);

	sprintf(fPath, "%s", f_Name); 
	if(labelChk == 'y' || labelChk == 'Y'){             

		labelData = (char **)malloc(size_x*sizeof(char*));

		if(GetData(fPath, inputData, labelData, size_x))
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);

			errorExit(0);
		}	
	
	}else{                         

		if(GetData2(fPath, inputData, size_x)) 
		{
			printf("File open success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open success");
			writeLogFile(fPath, logMsg, 0);
		}
		else
		{ 
			printf("File open error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "File open error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);

			errorExit(0);
		}

	}//labelChk		

	//rnd seed
	srand(time(NULL));	
	i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, h = 0, t = 0, x = 0, g = 0;
    startTime = 0l, endTime = 0l;  
		
	startTime = clock();
	///////////////////////////////////////////////////////////////////////
	///////////Data Decomposition Part
	///////////////////////////////////////////////////////////////////////
    
	for(i = 0; i < size_x; i++)
	{
		if(g == (p-1)){
			if(i % dSize_x_l == 0){
				divDataIdx[i] = (i * size_x) + i;			
				diag = divDataIdx[i];
			}else{
				divDataIdx[i] = diag + size_x;
				diag = divDataIdx[i];
			}
		}else{
			if(i % dSize_x == 0){
				divDataIdx[i] = (i * size_x) + i;			
				diag = divDataIdx[i];
			}else{
				divDataIdx[i] = diag + size_x;
				diag = divDataIdx[i];
			}
		}

		g++;
	}
	
	for(i = 0;i < p;i++)
	{
		diag = 0;

		if(i == (p-1)){
			t = dSize_x_l;
		}else{
			t = dSize_x;
		}

		for(j = 0; j < t * t; j++)
		{			
			if(diag == t)
			{
				diag = 0;
				idx++;
			}

			divData[i][j] = inputData[divDataIdx[idx] + diag];			
			
			diag++;			
		}
		
		idx++;
	}

	///////////////////////////////////////////////////////////////////////
	///////////Data Decomposition Part end
	///////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////
	///////////Data Sampling Part
	///////////////////////////////////////////////////////////////////////	
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Sampling process starts");
	writeLogFile(fPath, logMsg, 0);
    i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, h = 0, t = 0, x = 0, g = 0, idx = 0;    
	
	for(i = 0; i < p; i++)
	{
		if(i == (p-1)){
			s = s_l;
			g = dSize_x_l;
		}else{
			s = s_f;
			g = dSize_x;
		}

		t = 0;
		diag = rand() % g;                             //ù landmark
		samplePointChk[0] = diag;
		samplePointIdx[idx] = diag + (i*dSize_x);
		samplePointMDS[idx] = diag;
		idx++;

		for(k = 0; k < g; k++){
			minJ[k] = divData[i][(diag * g) + k];
			divDataS[k].rowIdx = diag;
			divDataS[k].colIdx = k;
			divDataS[k].value  = minJ[k];
		}

		for(j = 1; j < s; j++){				

			quicksortS(divDataS, 0, (g-1));

			for(h = 0; h < g; h++){				
				if((find_index(samplePointChk, t, divDataS[g - (h+1)].colIdx) < 0)){
					samplePointChk[j] = divDataS[g - (h+1)].colIdx;
					samplePointIdx[idx] = divDataS[g - (h+1)].colIdx + (i*dSize_x);
					samplePointMDS[idx] = divDataS[g - (h+1)].colIdx;
					idx++;t++;
					
					h = g;
				}//if						
			}//for h
			
			for(x = 0; x < g; x++)
			{
				if(minJ[x] > divData[i][(samplePointChk[j]*g) + x]){
					minJ[x] = divData[i][(samplePointChk[j]*g) + x];	

					divDataS[x].rowIdx = samplePointChk[j];
					divDataS[x].colIdx = x;
					divDataS[x].value  = minJ[x];
				}else{
					divDataS[x].rowIdx = samplePointChk[j-1];
					divDataS[x].colIdx = x;
					divDataS[x].value  = minJ[x];
				}
			}//for x

		}//for j

	}//for i

	sprintf(fPath, "%s_Indexes.txt", outName); 
	if(writeDataFileInt(fPath, mAlignSize, samplePointIdx)){
		printf("Index file write success\n");
	}else{
		printf("Index file write error\n");

		sprintf(fPath, "%s_Log.txt", outName);
		sprintf(logMsg, "Sampling process error");
		writeLogFile(fPath, logMsg, 0);
		sprintf(logMsg, "Program terminated");
		writeLogFile(fPath, logMsg, 0);
		writeLogFile(fPath, logMsg, 2);

		if(labelChk == 'y' || labelChk == 'Y'){
			free(labelData);
		}
		
		free(inputData);	
		free(samplePointIdx);
		free(samplePointMDS);	
		free(mAlign);
		free(mAlignTemp);
		free(mAlignMDS);	
		free(mAlignMDSTemp);
		free(divDataMDSTemp);
		free(affine);
		free(divDataS);
		free(divData);
		free(divDataIdx);
		free(divDataMDS);
		free(subMDScoor);
		free(MDScoor);
		free(samplePointChk);

		errorExit(4);
	}

	diag = 0; k = 1;idx = 0; mdsIdx = 0;
	for(i = 0; i < mAlignSize; i++)
	{
		for(j = i; j < mAlignSize; j++)
		{			
			if(i == j) 
			{
				mAlignTemp[i][j] = 0.0f;				
			}else{				
				mAlignTemp[i][j] = inputData[samplePointIdx[i] * size_x + samplePointIdx[j]];								
				mAlignTemp[j][i] = mAlignTemp[i][j];
			}   
		}
	}    

	idx = 0;
	for(i = 0; i < mAlignSize; i++)
	{
		for(j = 0; j < mAlignSize; j++)
		{
			mAlign[idx] = mAlignTemp[i][j];				
			idx++;
		}
	}        
    
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Sampling process success");
	writeLogFile(fPath, logMsg, 0);
	///////////////////////////////////////////////////////////////////////
	///////////Data Sampling end
	///////////////////////////////////////////////////////////////////////

	/* Initialize CUBLAS & CULA */
    status = cublasInit();
    cublasCheckStatus(status, params);
    cStatus = culaInitialize();
    checkStatus(cStatus, params);

    ///////////////////////////////////////////////////////////////////////
	///////////Classical MDS Part
	///////////////////////////////////////////////////////////////////////	
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "MDS process starts");
	writeLogFile(fPath, logMsg, 0);

	mAlignMDS = cMDS(mAlignSize, mAlignSize, dim, mAlign, params);                                
		
	for(i = 0; i<p; i++)
	{
		if(i == (p-1)){
			divDataMDS[i] = cMDS(dSize_x_l, dSize_x_l, dim, divData[i], params);         
		}else{
			divDataMDS[i] = cMDS(dSize_x, dSize_x, dim, divData[i], params);            
		}		
	}
	   
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "MDS process success");
	writeLogFile(fPath, logMsg, 0);
	///////////////////////////////////////////////////////////////////////
	///////////Classical MDS Part end
	///////////////////////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////////////////////
	///////////Calculate Least Linear Regression & Affine mapping Part
	///////////////////////////////////////////////////////////////////////
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Affine mapping process starts");
	writeLogFile(fPath, logMsg, 0);

    status = cublasAlloc(dSize_x*dim, sizeof(float), (void**)&d_data1);                                              //dMDS
    cublasCheckStatus(status, params);
    status = cublasAlloc(dim*dim, sizeof(float), (void**)&d_data2);                                                  //affine
    cublasCheckStatus(status, params);
    status = cublasAlloc(dSize_x*dim, sizeof(float), (void**)&d_result);                
    cublasCheckStatus(status, params);
	
	i = 0, j = 0, diag = 0, idx = 0, k = 0, mdsIdx = 0, h = 0, t = 0, x = 0, g = 0, idx = 0;
	for(diag = 0; diag<p; diag++)
	{
		if(diag == (p-1)){			
			s = s_l;			
			subMDScoor = (float *)malloc( dSize_x_l * dim * sizeof(float) );
			mAlignMDSTemp  = (float *)malloc(s_l*dim*sizeof(float));
			divDataMDSTemp = (float *)malloc(s_l*dim*sizeof(float));

			cublasFree(d_data1);
			cublasFree(d_data2);
			cublasFree(d_result);

			status = cublasAlloc(dSize_x_l*dim, sizeof(float), (void**)&d_data1);                                              //dMDS
			cublasCheckStatus(status, params);
			status = cublasAlloc(dim*dim, sizeof(float), (void**)&d_data2);                                                    //affine
			cublasCheckStatus(status, params);
			status = cublasAlloc(dSize_x_l*dim, sizeof(float), (void**)&d_result);                                         
			cublasCheckStatus(status, params);
			
			for(idx = 0; idx<dim; idx++){				
				for(k = 0; k<s_l; k++){
					if(idx == 0){
						t = (idx * s_f * p) + k + (diag * s_f);
					}else{
						t =  mAlignSize + k + (diag * s_f);
					}					
					mAlignMDSTemp[idx * s_l + k] = mAlignMDS[t];                                     
					divDataMDSTemp[idx * s_l + k] = divDataMDS[diag][idx * dSize_x_l + samplePointMDS[k + s_f * diag]];									
				}			
			}

			cStatus = culaSgels('N', s_l, dim, dim, divDataMDSTemp, s_l, mAlignMDSTemp, s_l);                                     
			checkStatus(cStatus, params);

			for( i = 0; i < dim; i++ ) {
				for( j = 0; j < dim; j++ ){
					affine[i*dim+j] = mAlignMDSTemp[i*s_l+j];
				}			
			}

			status = cublasSetVector(dSize_x_l*dim, sizeof(float), divDataMDS[diag], 1, d_data1, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dim*dim, sizeof(float), affine, 1, d_data2, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dSize_x_l*dim, sizeof(float), subMDScoor, 1, d_result, 1);
			cublasCheckStatus(status, params);
			cublasSgemm('n', 'n', dSize_x_l, dim, dim, 1.0f, d_data1, dSize_x_l, d_data2, dim, 0.0f, d_result, dSize_x_l);                 //dMDS * A
			cublasGetVector(dSize_x_l*dim, sizeof(float), d_result, 1, subMDScoor, 1);
						
			mdsIdx = 0;
			for(i = 0; i<dim; i++)
			{
				for( j = 0; j < dSize_x_l; j++ ){
					MDScoor[(i * size_x) + j + (diag * dSize_x)] = subMDScoor[i*dSize_x_l+j];
					mdsIdx++;
				}
			}

		}else{
		
			for(idx = 0; idx<dim; idx++){
				for(k = 0; k<s; k++){
					mAlignMDSTemp[idx * s + k] = mAlignMDS[(idx * s * p) + k + (diag * s)];                                    
					divDataMDSTemp[idx * s + k] = divDataMDS[diag][idx * dSize_x + samplePointMDS[k + s*diag]];				
				}			
			}

			cStatus = culaSgels('N', s, dim, dim, divDataMDSTemp, s, mAlignMDSTemp, s);                                        
			checkStatus(cStatus, params);

			for( i = 0; i < dim; i++ ) {
				for( j = 0; j < dim; j++ ){
					affine[i*dim+j] = mAlignMDSTemp[i*s+j];
				}			
			}

			status = cublasSetVector(dSize_x*dim, sizeof(float), divDataMDS[diag], 1, d_data1, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dim*dim, sizeof(float), affine, 1, d_data2, 1);
			cublasCheckStatus(status, params);
			status = cublasSetVector(dSize_x*dim, sizeof(float), subMDScoor, 1, d_result, 1);
			cublasCheckStatus(status, params);
			cublasSgemm('n', 'n', dSize_x, dim, dim, 1.0f, d_data1, dSize_x, d_data2, dim, 0.0f, d_result, dSize_x);                 //dMDS * A
			cublasGetVector(dSize_x*dim, sizeof(float), d_result, 1, subMDScoor, 1);
			
			//MDS ǥ combine
			mdsIdx = 0;
			for(i = 0; i<dim; i++)
			{
				for( j = 0; j < dSize_x; j++ ){
					MDScoor[(i * size_x) + j + (diag * dSize_x)] = subMDScoor[i*dSize_x+j];
					mdsIdx++;
				}
			}
		}//if(i == (p-1))

	}//diag
   
	sprintf(fPath, "%s_Log.txt", outName);
	sprintf(logMsg, "Affine mapping process success");
	writeLogFile(fPath, logMsg, 0);
	///////////////////////////////////////////////////////////////////////
	///////////Calculate Least Linear Regression Part
	///////////////////////////////////////////////////////////////////////

	sprintf(fPath, "%s_CFMDS.txt", outName); 	
	if(labelChk == 'y' || labelChk == 'Y'){
		if(writeResultFileH(fPath, size_x, dim, MDScoor, labelData)){
			printf("File write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);

			free(labelData);
		}else{
			printf("File write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			cublasFree(d_data1);	
			cublasFree(d_data2);
			cublasFree(d_result);
			
			free(labelData);
			free(inputData);	
			free(samplePointIdx);
			free(samplePointMDS);	
			free(mAlign);
			free(mAlignTemp);
			free(mAlignMDS);	
			free(mAlignMDSTemp);
			free(divDataMDSTemp);
			free(affine);
			free(divDataS);
			free(divData);
			free(divDataIdx);
			free(divDataMDS);
			free(subMDScoor);
			free(MDScoor);
			free(samplePointChk);

			culaShutdown();
			cublasShutdown();
			errorExit(4);
		}
	}else{
		if(writeResultFile(fPath, size_x, dim, MDScoor)){
			printf("File write success\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write success");
			writeLogFile(fPath, logMsg, 0);
		}else{
			printf("File write error\n");
			sprintf(fPath, "%s_Log.txt", outName);
			sprintf(logMsg, "Result file write error");
			writeLogFile(fPath, logMsg, 0);
			sprintf(logMsg, "Program terminated");
			writeLogFile(fPath, logMsg, 0);
			writeLogFile(fPath, logMsg, 2);

			cublasFree(d_data1);	
			cublasFree(d_data2);
			cublasFree(d_result);
						
			free(inputData);	
			free(samplePointIdx);
			free(samplePointMDS);	
			free(mAlign);
			free(mAlignTemp);
			free(mAlignMDS);	
			free(mAlignMDSTemp);
			free(divDataMDSTemp);
			free(affine);
			free(divDataS);
			free(divData);
			free(divDataIdx);
			free(divDataMDS);			
			free(subMDScoor);
			free(MDScoor);
			free(samplePointChk);

			culaShutdown();
			cublasShutdown();
			errorExit(4);
		}
	}//if(labelChk)
	
	endTime = clock();
	printf("Processing time : %ld (ms)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	sprintf(logMsg, "Processing time : %ld (sec)\n", (endTime - startTime)/CLOCKS_PER_SEC);
	writeLogFile(fPath, logMsg, 0);		
	sprintf(logMsg, "CFMDS Process success");
	writeLogFile(fPath, logMsg, 0);
	writeLogFile(fPath, logMsg, 2);
	
	cublasFree(d_data1);
	cublasFree(d_data2);
    cublasFree(d_result);

    free(inputData);	
    free(samplePointIdx);
	free(samplePointMDS);	
	free(mAlign);		
	free(mAlignTemp);
	free(mAlignMDS);	
	free(mAlignMDSTemp);
	free(divDataMDSTemp);
	free(affine);
	free(divDataS);
	free(divData);
	free(divDataIdx);
	free(divDataMDS);	
	free(subMDScoor);
	free(MDScoor);
	free(samplePointChk);		
	
	return 0;
}
